# This belongs somewhere else!
# Not used 

from scipy.optimize import newton

def force(x, x1, x2, m1, m2, expon=2):
    return m1 / (abs(x - x1) ** expon) - m2 / (abs(x - x2) ** 2)


def weighted_mid(x1, x2, m1, m2, expon=2):
    if x1 >= x2:
        return x1
    else:
        x_mid = 0.5*x1+0.5*x2
        x_weighted = x1 + m1/(m1+m2)**2 * (x2-x1)
        x_guess = 0.25*x_weighted+0.75*x_mid
        try:
            return newton(func=force, x0=x_guess, x1=x_mid, args=(x1, x2, m1, m2, expon), tol=(x2-x1)/100.0 )
        except:
            return x_guess

        
def iex_weighted_mid(d, expon=-1.5):
    x1 = d['iexBidPrice']
    x2 = d['iexAskPrice']
    m1 = d['iexBidSize']
    m2 = d['iexAskSize']
    if (x1 != 0) and (x2 != 0) and (m1 != 0) and (m2 != 0) and (x2 > x1):
        wm = weighted_mid(x1=x1, x2=x2, m1=m1, m2=m2)
        print({'x1': x1, 'm1': m1, 'x2': x2, 'm2': m2, 'wm': wm})
        return wm
    else:
        return d['latestPrice']




from microprediction.live.xraytickers import get_xray_tickers
from microprediction.live.iex import iex_latest_prices
import math


def get_xray_prices():
    """  Polling function
         Retrieve live prices for the tickers in the xray list
         api_key: An IEX api key
    """
    from microprediction.live.iexcredentials import get_iex_key
    IEX_KEY = get_iex_key()
    tickers = get_xray_tickers()
    sp = iex_latest_prices(tickers=tickers, api_key=IEX_KEY)
    return [1000 * math.log(v) for v in sp]



def player_classifier(ability):
     return {'great': abs(ability -1.67) < 0.33,
               'good': abs(ability -1.0) < 0.33,
               'okay': abs(ability -0.0) < 0.25,
               'bad': abs(ability+1.0) < 0.33}

# Strokes gained statistics
GOLF_SG_CATEGORIES = ['total','ott','app','arg','putt']



from pycoingecko import CoinGeckoAPI
from pprint import pprint

# New video tutorials are available at https://www.microprediction.com/python-1 to help you
# get started creating streams (see the 4th module in particular)

def names_and_prices():
    """ Returns bitcoin and ethereum in USD and EUR """
    # Just an example
    cg   = CoinGeckoAPI()
    data = cg.get_price(ids='bitcoin,ethereum', vs_currencies='usd,eur')
    # Flatten
    names  = list()
    prices = list()
    for coin, fx in data.items():
        for currency, price in fx.items():
            name  = coin+'_'+currency
            names.append(name)
            prices.append(price)

    return names, prices

if __name__=='__main__':
    names, values = names_and_prices()
    pprint(names)
    pprint(values)


try:
    from timemachines.skaters.tsa.tsaensembles import tsa_precision_combined_ensemble
    using_tsa = True
except ImportError:
    print('pip install --upgrade pip')
    print('pip install timemachines')
    using_tsa = False


import math
from microprediction.live.iex import iex_latest_prices


GRAIN_TICKERS = ['weat', 'corn', 'soyb']
GRAIN_NAMES = ['grains_' + ticker + '.json' for k,ticker in enumerate(GRAIN_TICKERS)]


def iex_scaled_log_grains(api_key:str):
    """
       Retrieve live prices for FAANG and transform
       api_key: An IEX api key
    """
    sp = iex_latest_prices(tickers=GRAIN_TICKERS, api_key=api_key)
    return [1000 * math.log(v) for v in sp]




import math
from microprediction.live.iex import iex_latest_prices, iex_common_stock_with_balance_sheet_tickers
import os
import json
from getjson import getjson
from microprediction.whereami import TOP
from microprediction.live.iex import iex_common_stock_with_balance_sheet_tickers


RDPS_DESCRIPTIONS = {'e':'energy',
         'c':'communication',
         'y':'discretionary',
         'p':'discretionary',
         'f':'financials',
         'v':'health',
         'i':'industrials',
         'b':'materials',
         're':'real estate',
         'k':'technology',
         'u':'utilities'}


RDPS_TICKERS = ['xl'+suffix for suffix in RDPS_DESCRIPTIONS.keys() ]

RDPS_GENERIC_NAMES = [ 'rdps_' + ticker + '.json' for ticker in RDPS_TICKERS ]

RDPS_TICKERS_JSON = os.path.join(TOP,'live', 'rdpstickers.json')
RDPS_TICKERS_REVERSE_JSON = os.path.join(TOP ,'live' ,'rdpstickersreverse.json')


def save_rdps_tickers(tickers):
    tickers_dict = dict([(k, ticker) for k, ticker in enumerate(tickers)])
    with open(RDPS_TICKERS_JSON, 'wt') as fp:
        json.dump(obj=tickers_dict, fp=fp)
    tickers_reverse_dict = dict(sorted([(ticker, k) for k, ticker in enumerate(tickers)]))
    with open(RDPS_TICKERS_REVERSE_JSON, 'wt') as fp:
        json.dump(obj=tickers_reverse_dict, fp=fp)


def get_rdps_tickers() -> [str]:
    """ Retrieve xray tickers """
    url = 'https://raw.githubusercontent.com/microprediction/microprediction/master/microprediction/live/rdpstickers.json'
    data = getjson(url, failover_url=url)
    return [data[str(i)] for i in range(len(data))]



if __name__=='__main__':
    save_rdps_tickers(RDPS_TICKERS)


from copy import deepcopy
import numpy as np
from collections import OrderedDict


class DistMachine(object):

    # This is a state engine that has a ppf function
    # Example of usage: SequentialStreamCrawler
    # Subclasses: LossDist, FitDist

    def __init__(self, state, params: OrderedDict = None):
        self.params = params
        self.state = state

    def update(self, value=None, dt=None, **kwargs):
        # Incorporate new value, time passing, or both
        # Typically will update the state but not params
        raise NotImplementedError

    def inv_cdf(self, p: float) -> float:
        # Something like StatsConventions.norminv(p)
        raise NotImplementedError


class LossDist(DistMachine):

    # A distribution machine with a loss function

    def __init__(self, state: dict, params: OrderedDict = None):
        super().__init__(state=state, params=params)

    def log_likelihood(self, value: float) -> float:
        """ Likelihood for one value """
        raise NotImplementedError

    # Likelihood Distribution Machine comes with a default loss function, though you could
    # choose to augment or override this by adding other penalties (for instance adding penalty
    # for moving or back-tracking too much)

    def loss(self, lagged_values, lagged_times, params=None, state=None):
        """ Negative log-likelihood """

        def pre_getter(state, value, machine):
            return machine.log_likelihood(value=value)

        lls = self.replay(pre_getter=pre_getter, lagged_values=lagged_values, lagged_times=lagged_times,
                          params=params, state=state)
        return -np.sum(lls)

    def replay(self, lagged_values, pre_getter=None, post_getter=None, lagged_times=None, params=None, state=None):
        """ Run through data, return whatever getter retrieves from state

                state         set initial state
                params        set params
                pre_getter    get_iex_realtime_price( state, value, machine=None ) -> any   Before .update() called
                post_getter   get_iex_realtime_price( state, value, machine=None ) -> any   After  .update() called
                returns: [ pre_getter(), post_getter(), pre_getter(),... ] but omits if None
        """
        saved_params = deepcopy(self.params)
        saved_state = deepcopy(self.state)
        if params is None:
            params = self.params
        if state is None:
            state = self.state

        self.params = deepcopy(params)
        self.state = deepcopy(state)

        chronological_values = list(reversed(lagged_values))
        chronological_times = list(reversed(lagged_times))
        chronological_dt = [1.0] + list(np.diff(chronological_times))

        accumulated = list()
        for value, dt in zip(chronological_values, chronological_dt):
            if pre_getter is not None:
                accumulated.append(pre_getter(state=self.state, value=value, machine=self))
            self.update(dt=dt)
            if post_getter is not None:
                accumulated.append(post_getter(state=self.state, value=value, machine=self))
            self.update(value=value)
        self.params = saved_params
        self.state = saved_state
        return accumulated



import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="microprediction",
    version="1.1.22",
    description="Client for www.microprediction.org turnkey community prediction",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/microprediction/microprediction",
    author="microprediction",
    author_email="info@microprediction.org",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    packages=["microprediction", "microprediction.live", "microprediction.inclusion","microprediction.univariate", "microprediction.bespoke",
              "microprediction.bespoke.meme_stocks","microprediction.bespoke.golf","microprediction.bespoke.crypto"],
    test_suite='pytest',
    tests_require=['pytest', 'scipy'],
    include_package_data=True,
    install_requires=["numpy>=1.20.1", "pandas", "contexttimer", "requests",
                      "getjson>=2.0.0", "microconventions>=1.0.0","pytz>=2021.3",
                      'pycoingecko', 'tdigest','genson','hyperopt',
                      'scikit-learn','statsmodels','copulas'],
    entry_points={
        "console_scripts": [
            "microprediction=microprediction.__main__:main",
        ]
    },
)



from microconventions.stats_conventions import is_discrete, evenly_spaced_percentiles, cdf_values,\
    is_process, nudged, discrete_cdf, sign_changes, quantize, StatsConventions
import numpy as np


# Utilities to help with CDF's

from microprediction.univariate.arrivals import approx_mode




# Illustrates a crawler that performs incremental estimation

# Video tutorials are available at https://www.microprediction.com/python-1 to help you
# get started. There's a video explanation of FitCrawler, SequentialCrawler and friends
# at https://www.microprediction.com/fitcrawler

from microprediction import MicroCrawler
import time
from collections import OrderedDict


class OnlineHorizonCrawler(MicroCrawler):

    #     # Crawler maintains a queue of horizons and cycles through them, calibrating as it goes
    #     # To use this crawler, override the initial_state() and update_state() methods
    #     # These store state keyed by horizon so your sample() method can make use of the state conveniently

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.queue = OrderedDict()  # Maintains an ordered list of state objects indexed by horizon

    ################################################################################
    #  Override these methods                                                      #
    ################################################################################

    def initial_state(self, name, delay=None, **ignore):
        """ How do you want state initialized? """
        return None

    def update_state(self, state, **ignore):
        return state

    ################################################################################
    #   Maybe don't override the methods below unless you know what you are doing  #
    ################################################################################

    def add_active_to_queue(self):
        """ Ensure there is a state object for each horizon """
        for horizon in self.get_active():
            if horizon not in self.queue:
                name, delay = self.split_horizon_name(horizon)
                self.queue.update({horizon: self.initial_state(name=name, delay=delay)})

    def downtime(self, seconds, **ignored):
        """ During downtime we cycle through the state and update it """
        self.add_active_to_queue()
        start_time = time.time()
        while time.time() < start_time + seconds - 5:
            if self.queue:
                state_key, state = self.queue.popitem(0)
                state = self.update_state(state)
                self.queue.update({state_key: state})
                time.sleep(3)


class OnlineStreamCrawler(OnlineHorizonCrawler):

    # TODO: Derive from StreamCrawler instead and present options to override
    #       sample() or alternatively, override sample_with_state()
    #       This is merely the same as StreamCrawler with a default downtime()

    # Crawler maintains a queue of streams and cycles through them, calibrating as it goes
    # To use this crawler, override the initial_state() and update_state() methods
    # These store state keyed by name so your sample() method can make use of the state conveniently

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.queue = OrderedDict()  # Maintains an ordered list of state objects indexed by stream name

    ################################################################################
    #  Override these methods                                                      #
    ################################################################################

    def initial_state(self, name, **ignore):
        """ How do you want state initialized? """
        return None

    def update_state(self, state, **ignore):
        return state

    ################################################################################
    #   Maybe don't override the methods below unless you know what you are doing  #
    ################################################################################

    def add_active_to_queue(self):
        """ Ensure there is a state object for each horizon """
        # You may want to change this to have a state object for each stream instead
        active_streams = [self.split_horizon_name(horizon)[0] for horizon in self.get_active()]
        for name in active_streams:
            if not name in self.queue:
                self.queue.update({name: self.initial_state(name)})
